import sys
from collections import namedtuple
from enum import IntEnum


class ProcessState(IntEnum):
    """Oredered enum to make state checks easy.

    For example Alive can be determined using >= ProcessState.ALIVE,
    which will return True if the state is READY as well as ALIVE.
    """
    NOT_STARTED = 0
    STARTED = 1
    ERROR = 2
    STOPPING = 3
    ALIVE = 4
    READY = 5


# Process state change callback mappings.
_STATUS_CALLBACKS = [
    'on_started',
    'on_alive',
    'on_ready',
    'on_error',
    'on_stopping',
]

# namedtuple defaults only available on 3.7 and later python versions
if sys.version_info < (3, 7):
    StatusCallbackMap = namedtuple('CallbackMap', _STATUS_CALLBACKS)
    StatusCallbackMap.__new__.__defaults__ = (None,) * 5
else:
    StatusCallbackMap = namedtuple(
        'CallbackMap',
        _STATUS_CALLBACKS,
        defaults=(None,) * len(_STATUS_CALLBACKS),
    )


class ProcessStatus:
    """Process status tracker.

    The class tracks process status and execute callback methods on
    state changes as well as replies to messagebus queries of the
    process status.

    Args:
        name (str): process name, will be used to create the messagebus
                    messagetype "mycroft.{name}...".
        bus (MessageBusClient): Connection to the Mycroft messagebus.
        callback_map (StatusCallbackMap): optionally, status callbacks for the
                                          various status changes.
    """

    def __init__(self, name, bus=None, callback_map=None, namespace="mycroft"):
        self.name = name
        self._namespace = namespace
        self.callbacks = callback_map or StatusCallbackMap()
        self.state = ProcessState.NOT_STARTED

        # Messagebus connection
        self.bus = None
        if bus:
            self.bind(bus)

    def bind(self, bus):
        self.bus = bus
        self._register_handlers()

    def _register_handlers(self):
        """Register messagebus handlers for status queries."""
        self.bus.on(f'{self._namespace}.{self.name}.is_alive', self.check_alive)
        self.bus.on(f'{self._namespace}.{self.name}.is_ready', self.check_ready)

        # The next one is for backwards compatibility
        self.bus.on(f'mycroft.{self.name}.all_loaded', self.check_ready)

    def check_alive(self, message=None):
        """Respond to is_alive status request.

        Args:
            message: Optional message to respond to, if omitted no message
                     is sent.

        Returns:
            bool, True if process is alive.
        """
        is_alive = self.state >= ProcessState.ALIVE

        if message:
            status = {'status': is_alive}
            self.bus.emit(message.response(data=status))

        return is_alive

    def check_ready(self, message=None):
        """Respond to all_loaded status request.

        Args:
            message: Optional message to respond to, if omitted no message
                     is sent.

        Returns:
            bool, True if process is ready.
        """
        is_ready = self.state >= ProcessState.READY
        if message:
            status = {'status': is_ready}
            self.bus.emit(message.response(data=status))

        return is_ready

    def set_started(self):
        """Process is started."""
        self.state = ProcessState.STARTED
        if self.callbacks.on_started:
            self.callbacks.on_started()

    def set_alive(self):
        """Basic loading is done."""
        self.state = ProcessState.ALIVE
        if self.callbacks.on_alive:
            self.callbacks.on_alive()

    def set_ready(self):
        """All loading is done."""
        self.state = ProcessState.READY
        if self.callbacks.on_ready:
            self.callbacks.on_ready()

    def set_stopping(self):
        """Process shutdown has started."""
        self.state = ProcessState.STOPPING
        if self.callbacks.on_stopping:
            self.callbacks.on_stopping()

    def set_error(self, err=''):
        """An error has occured and the process is non-functional."""
        # Intentionally leave is_started True
        self.state = ProcessState.ERROR
        if self.callbacks.on_error:
            self.callbacks.on_error(err)



from os.path import join, isdir, dirname, expanduser, isfile
from os import makedirs, listdir
from typing import List, Optional
from ovos_config.locations import get_xdg_data_save_path, get_xdg_data_dirs
from ovos_config.config import Configuration
from ovos_utils.log import LOG


def get_installed_skill_ids(conf: Optional[dict] = None) -> List[str]:
    """
    Gets a list of `skill_id`s for all installed skills
    Args:
        conf: Configuration, else loads from ovos_config.config.Configuration
    Returns:
        list of `skill_id` strings for all installed skills
    """
    _, skill_ids = get_plugin_skills()
    for d in get_skill_directories(conf):
        for skill_dir in listdir(d):
            if isdir(join(d, skill_dir)) and isfile(join(d, skill_dir,
                                                         "__init__.py")):
                if skill_dir in skill_ids:
                    LOG.info(f"{skill_dir} installed as plugin and local dir")
                    continue
                skill_ids.append(skill_dir)
    return skill_ids


def get_skill_directories(conf: Optional[dict] = None) -> List[str]:
    """ returns list of skill directories ordered by expected loading order
    This corresponds to:
    - XDG_DATA_DIRS
    - default directory (see get_default_skills_directory method for details)
    - user defined extra directories
    Each directory contains individual skill folders to be loaded
    If a skill exists in more than one directory (same folder name) previous instances will be ignored
        ie. directories at the end of the list have priority over earlier directories
    NOTE: empty folders are interpreted as disabled skills
    new directories can be defined in mycroft.conf by specifying a full path
    each extra directory is expected to contain individual skill folders to be loaded
    the xdg folder name can also be changed, it defaults to "skills"
        eg. ~/.local/share/mycroft/FOLDER_NAME
    {
        "skills": {
            "directory": "skills",
            "extra_directories": ["path/to/extra/dir/to/scan/for/skills"]
        }
    }
    Args:
        conf: Configuration, else loads from ovos_config.config.Configuration
    Returns:
        list of fully-qualified directories containing non-plugin skills
    """
    # the contents of each skills directory must be individual skill folders
    # we are still dependent on the mycroft-core structure of skill_id/__init__.py

    conf = conf or Configuration()

    # load all valid XDG paths
    # NOTE: skills are actually code, but treated as user data!
    # they should be considered applets rather than full applications
    skill_locations = list(reversed(
        [join(p, "skills") for p in get_xdg_data_dirs() if
         isdir(join(p, "skills"))]
    ))

    # load the default skills folder
    # only meaningful if xdg support is disabled
    default = get_default_skills_directory(conf)
    if default not in skill_locations:
        skill_locations.append(default)

    # load additional explicitly configured directories
    conf = conf.get("skills") or {}
    # extra_directories is a list of directories containing skill subdirectories
    # NOT a list of individual skill folders
    # preserve order while removing any duplicate entries
    extra_dirs = (expanduser(d) for d in conf.get("extra_directories") or [])
    for d in extra_dirs:
        if isdir(d) and d not in skill_locations:
            skill_locations.append(d)
    return skill_locations


def get_default_skills_directory(conf: Optional[dict] = None) -> str:
    """ return default directory to scan for skills
    This is only meaningful if xdg is disabled in ovos.conf
    If xdg is enabled then data_dir is always XDG_DATA_DIR
    If xdg is disabled then data_dir by default corresponds to /opt/mycroft
    users can define the data directory in mycroft.conf
    the skills folder name (relative to data_dir) can also be defined there
    NOTE: folder name also impacts all XDG skill directories!
    {
        "data_dir": "/opt/mycroft",
        "skills": {
            "directory": "skills"
        }
    }
    Args:
        conf: Configuration, else loads from ovos_config.config.Configuration
    Returns:
        Absolute path to default skills directory
    """
    conf = conf or Configuration()
    path_override = conf["skills"].get("directory_override")

    # if .conf wants to use a specific path, use it!
    if path_override:
        LOG.warning("'directory_override' is deprecated!\n"
                    "It will no longer be supported after version 0.0.3\n"
                    "add the new path to 'extra_directories' instead")
        skills_folder = expanduser(path_override)
    elif conf["skills"].get("extra_directories") and \
            len(conf["skills"].get("extra_directories")) > 0:
        skills_folder = expanduser(conf["skills"]["extra_directories"][0])
    else:
        skills_folder = join(get_xdg_data_save_path(), "skills")
    # create folder if needed
    try:
        makedirs(skills_folder, exist_ok=True)
    except PermissionError:  # old style /opt/mycroft/skills not available
        skills_folder = join(get_xdg_data_save_path(), "skills")
        makedirs(skills_folder, exist_ok=True)

    return skills_folder


def get_plugin_skills() -> (list, list):
    """
    Get the package directories for any pip installed skill plugins
    Returns:
        lists of skill directories and plugin skill IDs
    """
    import importlib.util
    try:
        from ovos_plugin_manager.skills import find_skill_plugins
    except ImportError:
        LOG.warning("ovos-plugin-manager not available to load plugin skills")
        return [], []
    skill_dirs = list()
    plugins = find_skill_plugins()
    skill_ids = list(plugins.keys())
    for skill_class in plugins.values():
        skill_dir = dirname(importlib.util.find_spec(
            skill_class.__module__).origin)
        skill_dirs.append(skill_dir)
    LOG.info(f"Located plugin skills: {skill_ids}")
    return skill_dirs, skill_ids



import requests
from json_database import JsonStorageXDG


class OVOSApiService:

    def __init__(self) -> None:
        self.uuid_storage = JsonStorageXDG("ovos_api_uuid")
        self.token_storage = JsonStorageXDG("ovos_api_token")

    def register_device(self):
        if self.check_if_uuid_exists():
            return
        else:
            created_challenge = requests.get('https://api.openvoiceos.com/create_challenge')
            challenge_response = created_challenge.json()
            register_device = requests.get(
                'https://api.openvoiceos.com/register_device/' + challenge_response['challenge'] + '/' +
                challenge_response['secret'])
            register_device_uuid = challenge_response['challenge']
            self.uuid_storage['uuid'] = register_device_uuid
            self.uuid_storage.store()

    def check_if_uuid_exists(self):
        if "uuid" in self.uuid_storage:
            return True
        return False

    def get_session_challenge(self):
        session_challenge_request = requests.get('https://api.openvoiceos.com/get_session_challenge')
        session_challenge_response = session_challenge_request.json()
        self.token_storage["challenge"] = session_challenge_response['challenge']
        self.token_storage.store()

    def get_uuid(self):
        return self.uuid_storage.get("uuid", "")

    def get_session_token(self):
        return self.token_storage.get("challenge", "")


class OvosWeather:
    def __init__(self):
        self.api = OVOSApiService()

    @property
    def uuid(self):
        return self.api.get_uuid()

    @property
    def headers(self):
        self.api.get_session_challenge()
        return {'session_challenge': self.api.get_session_token(), 'backend': 'OWM'}

    def get_current(self, query):
        reqdata = {"lat": query.get("lat"),
                   "lon": query.get("lon"),
                   "units": query.get("units"),
                   "lang": query.get("lang")}
        url = f"https://api.openvoiceos.com/weather/generate_current_weather_report/{self.uuid}"
        r = requests.post(url, data=reqdata, headers=self.headers)
        return r.json()

    def get_hourly(self, query):
        reqdata = {"lat": query.get("lat"),
                   "lon": query.get("lon"),
                   "units": query.get("units"),
                   "lang": query.get("lang")}
        url = f"https://api.openvoiceos.com/weather/generate_hourly_weather_report/{self.uuid}"
        r = requests.post(url, data=reqdata, headers=self.headers)
        return r.json()

    def get_forecast(self, query):
        # Requires Paid API
        reqdata = {"lat": query.get("lat"),
                   "lon": query.get("lon"),
                   "units": query.get("units"),
                   "lang": query.get("lang")}
        url = f"https://api.openvoiceos.com/weather/generate_forecast_weather_report/{self.uuid}"
        r = requests.post(url, data=reqdata, headers=self.headers)
        return r.json()

    def get_weather_onecall(self, query):
        reqdata = {"lat": query.get("lat"),
                   "lon": query.get("lon"),
                   "units": query.get("units"),
                   "lang": query.get("lang")}
        url = f'https://api.openvoiceos.com/weather/onecall_weather_report/{self.uuid}'
        r = requests.post(url, data=reqdata, headers=self.headers)
        return r.json()


class OvosWolframAlpha:
    def __init__(self):
        self.api = OVOSApiService()

    @property
    def uuid(self):
        return self.api.get_uuid()

    @property
    def headers(self):
        self.api.get_session_challenge()
        return {'session_challenge': self.api.get_session_token()}

    def get_wolfram_spoken(self, query):
        reqdata = {"input": query.get("input"),
                   "units": query.get("units")}
        url = f'https://api.openvoiceos.com/wolframalpha/spoken/{self.uuid}'
        r = requests.post(url, data=reqdata, headers=self.headers)
        return r

    def get_wolfram_simple(self, query):
        reqdata = {"input": query.get("input"),
                   "units": query.get("units")}
        url = f'https://api.openvoiceos.com/wolframalpha/simple/{self.uuid}'
        r = requests.post(url, data=reqdata, headers=self.headers)
        return r

    def get_wolfram_full(self, query):
        reqdata = {"input": query.get("input"),
                   "units": query.get("units"),
                   "output": query.get("output", "json")}
        url = f'https://api.openvoiceos.com/wolframalpha/full/{self.uuid}'
        r = requests.post(url, data=reqdata, headers=self.headers)
        if reqdata["output"] == "json":
            return r.json()
        else:
            return r


class OvosEdamamRecipe:
    def __init__(self):
        self.api = OVOSApiService()

    @property
    def uuid(self):
        return self.api.get_uuid()

    @property
    def headers(self):
        self.api.get_session_challenge()
        return {'session_challenge': self.api.get_session_token()}

    def get_recipe(self, query):
        reqdata = {"query": query.get("query"),
                   "count": query.get("count", 5)}
        url = f'https://api.openvoiceos.com/recipes/search_recipe/'
        r = requests.post(url, data=reqdata, headers=self.headers)
        return r.json()


class OvosOmdb:
    def __init__(self):
        self.api = OVOSApiService()

    @property
    def uuid(self):
        return self.api.get_uuid()

    @property
    def headers(self):
        self.api.get_session_challenge()
        return {'session_challenge': self.api.get_session_token()}

    def search_movie(self, query):
        reqdata = {"movie_name": query.get("movie_name"),
                   "movie_year": query.get("movie_year"),
                   "movie_id": query.get("movie_id")}

        url = f'https://api.openvoiceos.com/omdb/search_movie/'
        r = requests.post(url, data=reqdata, headers=self.headers)
        return r.json()



import subprocess
from distutils.spawn import find_executable


class InputDeviceHelper:
    def __init__(self) -> None:
        self.input_devices_list = []
        if not find_executable("libinput"):
            raise Exception("libinput executable not found")

    # ToDo: add support for discovring the input device based of a connected
    # monitors, currently linux only supports input listing directly from the
    # system

    def build_input_devices_list(self):
        # Always clear the list before building it
        self.input_devices_list.clear()

        input_device_names = []
        input_device_kernel_path = []
        input_device_group = []
        input_device_capabilities = []

        # Get the list of devices from libinput
        proc_output = subprocess.check_output(
            ['libinput', 'list-devices']).decode('utf-8')
        for line in proc_output.splitlines():
            if line.startswith('Device'):
                unformated_device_name = line.split(':')[1]
                unformated_device_name = unformated_device_name.strip()
                input_device_names.append(unformated_device_name)

        for line in proc_output.splitlines():
            if line.startswith('Kernel'):
                unformated_device_kernel_path = line.split(':')[1]
                unformated_device_kernel_path = unformated_device_kernel_path.strip()
                input_device_kernel_path.append(unformated_device_kernel_path)

        for line in proc_output.splitlines():
            if line.startswith('Group'):
                unformated_device_group = line.split(':')[1]
                unformated_device_group = unformated_device_group.strip()
                input_device_group.append(unformated_device_group)

        for line in proc_output.splitlines():
            if line.startswith("Capabilities"):
                unformated_device_capabilities = line.split(':')[1]
                unformated_device_capabilities = unformated_device_capabilities.strip()
                # check if there is a comma in the string and if so, split it
                if ' ' in unformated_device_capabilities:
                    unformated_device_capabilities = unformated_device_capabilities.split(
                        ' ')
                else:
                    unformated_device_capabilities = [
                        unformated_device_capabilities]

                input_device_capabilities.append(
                    unformated_device_capabilities)

        for i in range(len(input_device_names)):
            self.input_devices_list.append(
                {"Device": input_device_names[i],
                 "Kernel": input_device_kernel_path[i],
                 "Group": input_device_group[i],
                 "Capabilities": input_device_capabilities[i]
                 })

    def get_input_devices_list(self):
        self.build_input_devices_list()
        return self.input_devices_list

    def can_use_touch_mouse(self):
        # check if any of the devices support touch or mouse
        self.build_input_devices_list()
        for device in self.input_devices_list:
            if "touch" in device["Capabilities"] or "mouse" in device[
                    "Capabilities"] or "tablet" in device["Capabilities"] or "gesture" in device["Capabilities"]:
                return True

        return False

    def can_use_keyboard(self):
        self.build_input_devices_list()
        for device in self.input_devices_list:
            if "keyboard" in device["Capabilities"]:
                return True

        return False


def can_use_touch_mouse():
    return InputDeviceHelper().can_use_touch_mouse()


def can_use_keyboard():
    return InputDeviceHelper().can_use_keyboard()



from ovos_utils.enclosure.mark1 import Mark1EnclosureAPI
from ovos_utils import create_loop
from ovos_utils.log import LOG
from ovos_utils.messagebus import get_mycroft_bus
import random
from time import sleep
import collections
import copy


class FaceplateGrid(collections.MutableSequence):
    encoded = None
    str_grid = None
    pad_char = "."

    def __init__(self, grid=None, bus=None):
        self.bus = bus or get_mycroft_bus()
        self._api = Mark1EnclosureAPI(self.bus)
        self.grid = []
        for x in range(8):
            self.grid.append([])
            for y in range(32):
                self.grid[x].append(0)
        if self.encoded:
            self.grid = self.decode(self.encoded).grid
        elif self.str_grid is not None:
            self.grid = FaceplateGrid(bus=self.bus)\
                .from_string(self.str_grid).grid
        elif grid is not None:
            self.grid = grid

    @property
    def height(self):
        return len(self.grid)

    @property
    def width(self):
        return max([len(r) for r in self.grid])

    def display(self, invert=True, clear=True, x_offset=0, y_offset=0):
        self._api.mouth_display(self.encode(invert),
                                x_offset, y_offset, clear)

    def print(self, draw_padding=True, invert=False):
        print(self.to_string(draw_padding=draw_padding, invert=invert))

    def encode(self, invert=False):
        # to understand how this function works you need to understand how the
        # Mark I arduino proprietary encoding works to display to the faceplate

        # https://mycroft-ai.gitbook.io/docs/skill-development/displaying-information/mark-1-display

        # Each char value str_gridesents a width number starting with B=1
        # then increment 1 for the next. ie C=2
        width_codes = ['B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L',
                       'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W',
                       'X', 'Y', 'Z', '[', '\\', ']', '^', '_', '`', 'a']

        height_codes = ['B', 'C', 'D', 'E', 'F', 'G', 'H', 'I']

        encode = width_codes[self.width - 1]
        encode += height_codes[self.height - 1]

        # Turn the image pixels into binary values 1's and 0's
        # the Mark I face plate encoding uses binary values to
        # binary_values returns a list of 1's and 0s'. ie ['1', '1', '0', ...]
        binary_values = []
        for i in range(self.width):  # pixels
            for j in range(self.height):  # lines
                pixels = self.grid[j]

                if pixels[i] is None:  # padding
                    pixels[i] = 0

                if pixels[i] != 0:
                    if invert is False:
                        binary_values.append('1')
                    else:
                        binary_values.append('0')
                else:
                    if invert is False:
                        binary_values.append('0')
                    else:
                        binary_values.append('1')
        # these values are used to determine how binary values
        # needs to be grouped together
        number_of_bottom_pixel = 0

        if self.height > 4:
            number_of_top_pixel = 4
            number_of_bottom_pixel = self.height - 4
        else:
            number_of_top_pixel = self.height

        # this loop will group together the individual binary values
        # ie. binary_list = ['1111', '001', '0101', '100']
        binary_list = []
        binary_code = ''
        increment = 0
        alternate = False
        for val in binary_values:
            binary_code += val
            increment += 1
            if increment == number_of_top_pixel and alternate is False:
                # binary code is reversed for encoding
                binary_list.append(binary_code[::-1])
                increment = 0
                binary_code = ''
                alternate = True
            elif increment == number_of_bottom_pixel and alternate is True:
                binary_list.append(binary_code[::-1])
                increment = 0
                binary_code = ''
                alternate = False
        # Code to let the Mark I arduino know where to place the
        # pixels on the faceplate
        pixel_codes = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
                       'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P']
        for binary_values in binary_list:
            number = int(binary_values, 2)
            pixel_code = pixel_codes[number]
            encode += pixel_code
        return encode

    def decode(self, encoded, invert=False, pad=True):
        codes = list(encoded)

        # Each char value str_gridesents a width number starting with B=1
        # then increment 1 for the next. ie C=2
        width_codes = ['B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L',
                       'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W',
                       'X', 'Y', 'Z', '[', '\\', ']', '^', '_', '`', 'a']

        height_codes = ['B', 'C', 'D', 'E', 'F', 'G', 'H', 'I']

        height = height_codes.index(codes[1]) + 1
        width = width_codes.index(codes[0]) + 1

        # Code to let the Mark I arduino know where to place the
        # pixels on the faceplate
        pixel_codes = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
                       'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P']
        codes.reverse()
        binary_list = []
        for pixel_code in codes[:-2]:
            number = pixel_codes.index(pixel_code.upper())
            bin_str = str(bin(number))[2:]
            while not len(bin_str) == 4:
                bin_str = "0" + bin_str
            binary_list += [bin_str]

        binary_list.reverse()

        for idx, binary_code in enumerate(binary_list):
            # binary code is reversed for encoding
            binary_list[idx] = binary_code[::-1]

        binary_code = "".join(binary_list)

        # Turn the image pixels into binary values 1's and 0's
        # the Mark I face plate encoding uses binary values to
        # binary_values returns a list of 1's and 0s'. ie ['1', '1', '0', ...]
        grid = []
        # binary_code is a sequence of column by column
        cols = [list(binary_code)[x:x + height] for x in
                range(0, len(list(binary_code)), height)]

        for x in range(height):
            row = []
            for y in range(width):
                bit = int(cols[y][x])
                if invert:
                    if bit:
                        bit = 0
                    else:
                        bit = 1
                row.append(bit)
            grid.append(row)

        #  handle padding
        if pad:
            if width < self.width:
                n = int((self.width - width) / 2)
                if invert:
                    padding = [1] * n
                else:
                    padding = [0] * n
                for idx, row in enumerate(grid):
                    grid[idx] = padding + row + padding
            if height < self.height:
                pass # TODO vertical padding
        self.grid = grid
        return self

    def from_string(self, str_grid):
        rows = [r for r in str_grid.split("\n") if len(r)]
        grid = []
        for r in rows:
            row = []
            for char in list(r):
                if char == " ":
                    row.append(1)
                elif char == FaceplateGrid.pad_char:
                    row.append(None)
                else:
                    row.append(0)
            while len(row) < self.width:
                row.append(None)
            grid.append(row)
        self.grid = grid
        return self

    def to_string(self, draw_padding=False, invert=False):
        str_grid = ""
        for row in self.grid:
            line = ""
            for col in row:
                if col is None and draw_padding:
                    line += self.pad_char
                elif col == 1:
                    if invert:
                        line += "X"
                    else:
                        line += " "
                elif col == 0:
                    if invert:
                        line += " "
                    else:
                        line += "X"
            str_grid += line + "\n"
        return str_grid

    def invert(self):
        for x in range(self.height):
            for y in range(self.width):
                if self.grid[x][y] == 0:
                    self.grid[x][y] = 1
                elif self.grid[x][y] == 1:
                    self.grid[x][y] = 0
        return self

    def clear(self):
        for x in range(self.height):
            for y in range(self.width):
                self.grid[x][y] = 0
        return self

    @property
    def is_empty(self):
        for x in range(self.height):
            for y in range(self.width):
                if self.grid[x][y] == 1:
                    return False
        return True

    def randomize(self, n=200):
        for i in range(n):
            x = random.randint(0, self.height-1)
            y = random.randint(0, self.width-1)
            self.grid[x][y] = int(random.randint(0, 1))
        return self

    def __len__(self):
        # number of pixels
        return self.width * self.height

    def __delitem__(self, index):
        self.grid.__delitem__(index)

    def insert(self, index, value):
        self.grid.insert(index - 1, value)

    def __setitem__(self, index, value):
        self.grid.__setitem__(index, value)

    def __getitem__(self, index):
        return self.grid.__getitem__(index)


class FacePlateAnimation(FaceplateGrid):

    def __init__(self, grid=None, bus=None):
        super().__init__(grid, bus)
        self.finished = False

    def animate(self):
        pass

    def __iter__(self):
        while not self.finished:
            self.animate()
            yield self

    def start(self):
        self.finished = False

    def stop(self):
        self.finished = True

    def run(self, delay=0.5, callback=None, daemonic=False):
        self.start()

        if delay < 0.4:
            # writer bugs out if sending messages too rapidly
            delay = 0.4

        def step(callback=callback):
            try:
                if not self.finished:
                    self.animate()
                    if callback:
                        callback(self)
            except Exception as e:
                LOG.error(e)

        if daemonic:
            create_loop(step, delay)
        else:
            while not self.finished:
                step()
                sleep(delay)
        self.stop()

    def scroll_down(self):
        old = copy.deepcopy(self.grid)
        for x in range(self.width):
            for y in range(self.height):
                self.grid[y][x] = old[y - 1][x]

    def scroll_up(self):
        old = copy.deepcopy(self.grid)
        for x in range(self.width):
            for y in range(self.height):
                if y == self.height - 1:
                    self.grid[y][x] = old[0][x]
                else:
                    self.grid[y][x] = old[y + 1][x]

    def scroll_right(self):
        old = copy.deepcopy(self.grid)
        for x in range(self.width):
            for y in range(self.height):
                self.grid[y][x] = old[y][x - 1]

    def scroll_left(self):
        old = copy.deepcopy(self.grid)
        for x in range(self.width):
            for y in range(self.height):
                if x == self.width -1:
                    self.grid[y][x] = old[y][0]
                else:
                    self.grid[y][x] = old[y][x + 1]

    def move_down(self):
        old = copy.deepcopy(self.grid)
        for x in range(self.width):
            for y in range(self.height):
                if y - 1 < 0:
                    self.grid[y][x] = 0
                else:
                    self.grid[y][x] = old[y - 1][x]

    def move_up(self):
        old = copy.deepcopy(self.grid)
        for x in range(self.width):
            for y in range(self.height):
                if y == self.height - 1:
                    self.grid[y][x] = 0
                else:
                    self.grid[y][x] = old[y + 1][x]

    def move_right(self):
        old = copy.deepcopy(self.grid)
        for x in range(self.width):
            for y in range(self.height):
                self.grid[y][x] = old[y][x - 1]

    def move_left(self):
        old = copy.deepcopy(self.grid)
        for x in range(self.width):
            for y in range(self.height):
                if x == self.width - 1:
                    self.grid[y][x] = 0
                else:
                    self.grid[y][x] = old[y][x + 1]


class BlackScreen(FaceplateGrid):
    # Basically a util class to handle
    # inverting on __init__
    str_grid = """
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.invert()



# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# This file is directly copied from HolmesV, it's a simple utility to
# interface with the AudioService via messagebus outside core
from os.path import abspath
from datetime import timedelta
from ovos_utils.messagebus import Message, get_mycroft_bus


def ensure_uri(s):
    """Interprete paths as file:// uri's.

    Args:
        s: string to be checked

    Returns:
        if s is uri, s is returned otherwise file:// is prepended
    """
    if isinstance(s, str):
        if '://' not in s:
            return 'file://' + abspath(s)
        else:
            return s
    elif isinstance(s, (tuple, list)):
        if '://' not in s[0]:
            return 'file://' + abspath(s[0]), s[1]
        else:
            return s
    else:
        raise ValueError('Invalid track')


class AudioServiceInterface:
    """AudioService class for interacting with the audio subsystem

    Args:
        bus: Mycroft messagebus connection
    """

    def __init__(self, bus=None):
        self.bus = bus or get_mycroft_bus()

    def queue(self, tracks=None):
        """Queue up a track to playing playlist.

        Args:
            tracks: track uri or list of track uri's
                    Each track can be added as a tuple with (uri, mime)
                    to give a hint of the mime type to the system
        """
        tracks = tracks or []
        if isinstance(tracks, (str, tuple)):
            tracks = [tracks]
        elif not isinstance(tracks, list):
            raise ValueError
        tracks = [ensure_uri(t) for t in tracks]
        self.bus.emit(Message('mycroft.audio.service.queue',
                              data={'tracks': tracks}))

    def play(self, tracks=None, utterance=None, repeat=None):
        """Start playback.

        Args:
            tracks: track uri or list of track uri's
                    Each track can be added as a tuple with (uri, mime)
                    to give a hint of the mime type to the system
            utterance: forward utterance for further processing by the
                        audio service.
            repeat: if the playback should be looped
        """
        repeat = repeat or False
        tracks = tracks or []
        utterance = utterance or ''
        if isinstance(tracks, (str, tuple)):
            tracks = [tracks]
        elif not isinstance(tracks, list):
            raise ValueError
        tracks = [ensure_uri(t) for t in tracks]
        self.bus.emit(Message('mycroft.audio.service.play',
                              data={'tracks': tracks,
                                    'utterance': utterance,
                                    'repeat': repeat}))

    def stop(self):
        """Stop the track."""
        self.bus.emit(Message('mycroft.audio.service.stop'))

    def next(self):
        """Change to next track."""
        self.bus.emit(Message('mycroft.audio.service.next'))

    def prev(self):
        """Change to previous track."""
        self.bus.emit(Message('mycroft.audio.service.prev'))

    def pause(self):
        """Pause playback."""
        self.bus.emit(Message('mycroft.audio.service.pause'))

    def resume(self):
        """Resume paused playback."""
        self.bus.emit(Message('mycroft.audio.service.resume'))

    def get_track_length(self):
        """
        getting the duration of the audio in seconds
        """
        length = 0
        info = self.bus.wait_for_response(
            Message('mycroft.audio.service.get_track_length'),
            timeout=1)
        if info:
            length = info.data.get("length") or 0
        return length / 1000  # convert to seconds

    def get_track_position(self):
        """
        get current position in seconds
        """
        pos = 0
        info = self.bus.wait_for_response(
            Message('mycroft.audio.service.get_track_position'),
            timeout=1)
        if info:
            pos = info.data.get("position") or 0
        return pos / 1000  # convert to seconds

    def set_track_position(self, seconds):
        """Seek X seconds.

        Arguments:
            seconds (int): number of seconds to seek, if negative rewind
        """
        self.bus.emit(Message('mycroft.audio.service.set_track_position',
                              {"position": seconds * 1000}))  # convert to ms

    def seek(self, seconds=1):
        """Seek X seconds.

        Args:
            seconds (int): number of seconds to seek, if negative rewind
        """
        if isinstance(seconds, timedelta):
            seconds = seconds.total_seconds()
        if seconds < 0:
            self.seek_backward(abs(seconds))
        else:
            self.seek_forward(seconds)

    def seek_forward(self, seconds=1):
        """Skip ahead X seconds.

        Args:
            seconds (int): number of seconds to skip
        """
        if isinstance(seconds, timedelta):
            seconds = seconds.total_seconds()
        self.bus.emit(Message('mycroft.audio.service.seek_forward',
                              {"seconds": seconds}))

    def seek_backward(self, seconds=1):
        """Rewind X seconds

         Args:
            seconds (int): number of seconds to rewind
        """
        if isinstance(seconds, timedelta):
            seconds = seconds.total_seconds()
        self.bus.emit(Message('mycroft.audio.service.seek_backward',
                              {"seconds": seconds}))

    def track_info(self):
        """Request information of current playing track.

        Returns:
            Dict with track info.
        """
        info = self.bus.wait_for_response(
            Message('mycroft.audio.service.track_info'),
            reply_type='mycroft.audio.service.track_info_reply',
            timeout=1)
        return info.data if info else {}

    def available_backends(self):
        """Return available audio backends.

        Returns:
            dict with backend names as keys
        """
        msg = Message('mycroft.audio.service.list_backends')
        response = self.bus.wait_for_response(msg)
        return response.data if response else {}

    @property
    def is_playing(self):
        """True if the audioservice is playing, else False."""
        return self.track_info() != {}



from ovos_utils.system import search_mycroft_core_location, is_running_from_module
from ovos_utils.xdg_utils import (
    xdg_config_home,
    xdg_config_dirs,
    xdg_data_home,
    xdg_data_dirs,
    xdg_cache_home
)
from ovos_utils.log import LOG

from ovos_config.locations import (
    get_xdg_config_dirs,
    get_xdg_data_dirs,
    get_xdg_data_save_path,
    get_xdg_config_save_path,
    get_xdg_cache_save_path,
    find_default_config,
    find_user_config,
    get_config_locations,
    get_webcache_location,
    get_xdg_config_locations
)

from ovos_config.locale import get_default_lang
from ovos_config.meta import (
    get_ovos_config,
    get_ovos_default_config_paths,
    is_using_xdg,
    get_xdg_base,
    set_xdg_base,
    set_config_filename,
    set_default_config,
    get_config_filename
)

from ovos_config.config import (
    read_mycroft_config,
    update_mycroft_config
)

from ovos_config.models import (
    LocalConf,
    ReadOnlyConfig,
    MycroftUserConfig,
    MycroftDefaultConfig,
    MycroftSystemConfig,
    MycroftXDGConfig
)

from ovos_config.meta import save_ovos_config as save_ovos_core_config

LOG.warning("configuration moved to the `ovos_config` package. This submodule "
            "will be removed in ovos_utils 0.1.0")


def set_config_name(name, core_folder=None):
    # TODO deprecate, was only out in a couple versions
    LOG.warning("This reference is deprecated, use "
                "`ovos_config.meta.set_config_filename`")
    # renamed to match HolmesV
    set_config_filename(name, core_folder)



from ovos_utils.enclosure.mark1.faceplate import FaceplateGrid, BlackScreen


# drawing in python


class MusicIcon(BlackScreen):
    str_grid = """
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXX     XXXXXXXXXXXXX
XXXXXXXXXXXXXX     XXXXXXXXXXXXX
XXXXXXXXXXXXXX XXX XXXXXXXXXXXXX
XXXXXXXXXXXXXX XXX XXXXXXXXXXXXX
XXXXXXXXXXXXX  XX  XXXXXXXXXXXXX
XXXXXXXXXXXX   X   XXXXXXXXXXXXX
XXXXXXXXXXXXX XXX XXXXXXXXXXXXXX
"""


class PlusIcon(BlackScreen):
    str_grid = """
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXX   XXXXXXXXXXXXXX
XXXXXXXXXXXXXXX   XXXXXXXXXXXXXX
XXXXXXXXXXXXX       XXXXXXXXXXXX
XXXXXXXXXXXXX       XXXXXXXXXXXX
XXXXXXXXXXXXX       XXXXXXXXXXXX
XXXXXXXXXXXXXXX   XXXXXXXXXXXXXX
XXXXXXXXXXXXXXX   XXXXXXXXXXXXXX
"""


class HeartIcon(FaceplateGrid):
    str_grid = """
                                
            xx   xx             
           xxxx xxxx            
           xxxxxxxxx            
            xxxxxxx             
             xxxxx              
              xxx               
               x                
"""


class HollowHeartIcon(FaceplateGrid):
    str_grid = """
                                
            xx   xx             
           x  x x  x            
           x   x   x            
            x     x             
             x   x              
              x x               
               x                
"""


class SkullIcon(FaceplateGrid):
    str_grid = """
                                
            xxxxxxx             
           x  xxx  x            
           xxxxxxxxx            
            xxx xxx             
             xxxxx              
             x x x              
             x x x              
"""


class DeadFishIcon(FaceplateGrid):
    str_grid = """
                                
        x          xxxx         
         x  x x x xx xxx        
          xxxxxxxxxxxxxxx       
         x  x x x xxxxxx        
        x          xxxx         
"""


class InfoIcon(BlackScreen):
    str_grid = """
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXXX   XXXXXXXXXXXXXX
XXXXXXXXXXXXXXX   XXXXXXXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXXX    XXXXXXXXXXXXXX
XXXXXXXXXXXXXXX   XXXXXXXXXXXXXX
XXXXXXXXXXXXXXX   XXXXXXXXXXXXXX
XXXXXXXXXXXXXXX    XXXXXXXXXXXXX
"""


class ArrowLeftIcon(BlackScreen):
    str_grid = """
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXX   XXXXXXXXXXXXXXXX
XXXXXXXXXXXX   XXXXXXXXXXXXXXXXX
XXXXXXXXXXX   X       XXXXXXXXXX
XXXXXXXXXX   X        XXXXXXXXXX
XXXXXXXXXXX   X       XXXXXXXXXX
XXXXXXXXXXXX   XXXXXXXXXXXXXXXXX
XXXXXXXXXXXXX   XXXXXXXXXXXXXXXX
"""


class WarningIcon(BlackScreen):
    str_grid = """
XXXXXXXXXXXXXXX XXXXXXXXXXXXXXXX
XXXXXXXXXXXXXX   XXXXXXXXXXXXXXX
XXXXXXXXXXXXX  X  XXXXXXXXXXXXXX
XXXXXXXXXXXX  XXX  XXXXXXXXXXXXX
XXXXXXXXXXX   XXX   XXXXXXXXXXXX
XXXXXXXXXX           XXXXXXXXXXX
XXXXXXXXX     XXX     XXXXXXXXXX
XXXXXXXX       X       XXXXXXXXX
"""


class CrossIcon(BlackScreen):
    str_grid = """
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXXXX XXXXX XXXXXXXXXXXX
XXXXXXXXXXXX   XXX   XXXXXXXXXXX
XXXXXXXXXXXXX   X   XXXXXXXXXXXX
XXXXXXXXXXXXXXX   XXXXXXXXXXXXXX
XXXXXXXXXXXXX   X   XXXXXXXXXXXX
XXXXXXXXXXXX   XXX   XXXXXXXXXXX
XXXXXXXXXXXXX XXXXX XXXXXXXXXXXX
"""


class JarbasAI(BlackScreen):
    str_grid = """
X   XXXXXXXXXXXXXXXXXXXXXXX   X 
XX XXXXXXXXXXXXXXXXXXXXXXXX X XX
XX XXXXXXXXXX XXXXXXXXXXXXX X X 
XX XXXXXXXXXX XXXXXXXXX   X   X 
XX XX   X X X    XX   X XXX X X 
 X X XX X  XX XX X XX X   X X X 
 X X XX X XXX XX X XX XXX X X X 
   X    X XXX    X    X   X X X 
"""


class SpaceInvader1(BlackScreen):
    str_grid = """
XXXXXXXXXXXXXX     XXXXXXXXXXXXX
XXXXXXXXXXXXX       XXXXXXXXXXXX
XXXXXXXXXX     x x     XXXXXXXXX
XXXXXXXXXX XX       XX XXXXXXXXX
XXXXXXXXXXXXXX     XXXXXXXXXXXXX
XXXXXXXXXX             XXXXXXXXX
XXXXXXXXXX XXXXXXXXXXX XXXXXXXXX
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
"""


class SpaceInvader2(BlackScreen):
    str_grid = """
XXXXXXXXXXXXX       XXXXXXXXXXXX
XXXXXXXXXXXX   x x   XXXXXXXXXXX
XXXXXXXXXXXXX       XXXXXXXXXXXX
XXXXXXXXXX X         X XXXXXXXXX
XXXXXXXXXX             XXXXXXXXX
XXXXXXXXXX XXX     XXX XXXXXXXXX
XXXXXXXXXXXXXXX X XXXXXXXXXXXXXX
XXXXXXXXXXXXXXX X XXXXXXXXXXXXXX
"""


class SpaceInvader3(BlackScreen):
    str_grid = """
XXXXXXXXXXXXX       XXXXXXXXXXXX
XXXXXXXXXXXXXXXX XXXXXXXXXXXXXXX
XXXXXXXXXXXXX        XXXXXXXXXXX
XXXXXXXXXXXX   X X   XXXXXXXXXXX
XXXXXXXXXX             XXXXXXXXX
XXXXXXXXXX XX  XXX  XX XXXXXXXXX
XXXXXXXXXXXXXX     XXXXXXXXXXXXX
XXXXXXXXXXXXX  XXX  XXXXXXXXXXXX
"""


class SpaceInvader4(BlackScreen):
    str_grid = """
XXXXXXXXXXXXX       XXXXXXXXXXXX
XXXXXXXXXXXXXXX   XXXXXXXXXXXXXX
XXXXXXXXXXXXX        XXXXXXXXXXX
XXXXXXXXXXXX    X    XXXXXXXXXXX
XXXXXXXXXX             XXXXXXXXX
XXXXXXXXXX XX       XX XXXXXXXXX
XXXXXXXXXXXXXX  X  XXXXXXXXXXXXX
XXXXXXXXXXXXX  XXX  XXXXXXXXXXXX
"""


# Encoded icons
class Boat(BlackScreen):
    encoded = "QIAAABACAGIEMEOEPHAEAGACABABAAAAAA"


# Default weather icons for mark1
class SunnyIcon(FaceplateGrid):
    encoded = "IICEIBMDNLMDIBCEAA"


class PartlyCloudyIcon(FaceplateGrid):
    encoded = "IIEEGBGDHLHDHBGEEA"


class CloudyIcon(FaceplateGrid):
    encoded = "IIIBMDMDODODODMDIB"


class LightRainIcon(FaceplateGrid):
    encoded = "IIMAOJOFPBPJPFOBMA"


class RainIcon(FaceplateGrid):
    encoded = "IIMIOFOBPFPDPJOFMA"


class StormIcon(FaceplateGrid):
    encoded = "IIAAIIMEODLBJAAAAA"


class SnowIcon(FaceplateGrid):
    encoded = "IIJEKCMBPHMBKCJEAA"


class WindIcon(FaceplateGrid):
    encoded = "IIABIBIBIJIJJGJAGA"




# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import inspect
import logging
import sys
from os.path import join
from logging.handlers import RotatingFileHandler
from mycroft_bus_client.message import dig_for_message


class LOG:
    """
    Custom logger class that acts like logging.Logger
    The logger name is automatically generated by the module of the caller

    Usage:
        >>> LOG.debug('My message: %s', debug_str)
        13:12:43.673 - :<module>:1 - DEBUG - My message: hi
        >>> LOG('custom_name').debug('Another message')
        13:13:10.462 - custom_name - DEBUG - Another message
    """
    base_path = "stdout"
    fmt = '%(asctime)s.%(msecs)03d - ' \
          '%(name)s - %(levelname)s - %(message)s'
    datefmt = '%Y-%m-%d %H:%M:%S'
    formatter = logging.Formatter(fmt, datefmt)
    max_bytes = 50000000
    backup_count = 3
    name = 'OVOS'
    level = "DEBUG"
    diagnostic_mode = False
    _loggers = {}

    @classmethod
    def init(cls, config=None):
        config = config or {}
        cls.base_path = config.get("path", "stdout")
        cls.max_bytes = config.get("max_bytes", 50000000)
        cls.backup_count = config.get("backup_count", 3)
        cls.level = config.get("level", "INFO")
        cls.diagnostic_mode = config.get("diagnostic", False)

    @classmethod
    def create_logger(cls, name, tostdout=False):
        if name in cls._loggers:
            return cls._loggers[name]
        logger = logging.getLogger(name)
        logger.propagate = False
        # also log to stdout
        if tostdout or cls.base_path == "stdout":
            stdout_handler = logging.StreamHandler(sys.stdout)
            stdout_handler.setFormatter(cls.formatter)
            logger.addHandler(stdout_handler)
        # log to file
        if cls.base_path != "stdout":
            path = join(cls.base_path,
                        cls.name.lower().strip() + ".log")
            handler = RotatingFileHandler(path, maxBytes=cls.max_bytes,
                                          backupCount=cls.backup_count)
            handler.setFormatter(cls.formatter)
            logger.addHandler(handler)
        logger.setLevel(cls.level)
        cls._loggers[name] = logger
        return logger

    @classmethod
    def set_level(cls, level):
        cls.level = level
        for l in cls._loggers:
            cls._loggers[l].setLevel(level)

    @classmethod
    def _get_real_logger(cls):
        name = ""
        if cls.name is not None:
            name = cls.name + " - "

        # Stack:
        # [0] - _log()
        # [1] - debug(), info(), warning(), or error()
        # [2] - caller
        stack = inspect.stack()

        # Record:
        # [0] - frame object
        # [1] - filename
        # [2] - line number
        # [3] - function
        # ...
        record = stack[2]
        mod = inspect.getmodule(record[0])
        module_name = mod.__name__ if mod else ''
        name += module_name + ':' + record[3] + ':' + str(record[2])

        logger = cls.create_logger(name)
        if cls.diagnostic_mode:
            msg = dig_for_message()
            if msg:
                logger.debug(f"DIAGNOSTIC - source bus message {msg.serialize()}")
        return logger

    @classmethod
    def info(cls, *args, **kwargs):
        cls._get_real_logger().info(*args, **kwargs)

    @classmethod
    def debug(cls, *args, **kwargs):
        cls._get_real_logger().debug(*args, **kwargs)

    @classmethod
    def warning(cls, *args, **kwargs):
        cls._get_real_logger().warning(*args, **kwargs)

    @classmethod
    def error(cls, *args, **kwargs):
        cls._get_real_logger().error(*args, **kwargs)

    @classmethod
    def exception(cls, *args, **kwargs):
        cls._get_real_logger().exception(*args, **kwargs)





from ovos_utils.messagebus import Message, get_mycroft_bus
from ovos_utils.log import LOG
from time import sleep


class IntentLayers:
    def __init__(self, bus=None, layers=None):
        layers = layers or []
        self.bus = bus or get_mycroft_bus()
        # make intent levels for N layers
        self.layers = layers
        self.current_layer = 0
        self.activate_layer(0)
        self.named_layers = {}

    def disable_intent(self, intent_name):
        """Disable a registered intent"""
        self.bus.emit(Message("mycroft.skill.disable_intent",
                              {"intent_name": intent_name}))

    def enable_intent(self, intent_name):
        """Reenable a registered self intent"""
        self.bus.emit(Message("mycroft.skill.enable_intent",
                              {"intent_name": intent_name}))

    def reset(self):
        LOG.info("Reseting Intent Layers")
        self.activate_layer(0)

    def next(self):
        LOG.info("Going to next Intent Layer")
        self.current_layer += 1
        if self.current_layer > len(self.layers):
            LOG.info("Already in last layer, going to layer 0")
            self.current_layer = 0
        self.activate_layer(self.current_layer)

    def previous(self):
        LOG.info("Going to previous Intent Layer")
        self.current_layer -= 1
        if self.current_layer < 0:
            self.current_layer = 0
            LOG.error("Already in layer 0")
        else:
            self.activate_layer(self.current_layer)

    def add_layer(self, intent_list=None):
        intent_list = intent_list or []
        self.layers.append(intent_list)
        LOG.info("Adding intent layer: " + str(intent_list))

    def add_named_layer(self, name, intent_list=None):
        intent_list = intent_list or []
        self.named_layers[name] = len(self.layers)
        self.layers.append(intent_list)
        LOG.info("Setting layer " + name + " to: " + str(intent_list))

    def activate_named_layer(self, name):
        if name in self.named_layers:
            i = self.named_layers[name]
            LOG.info("activating layer named: " + name)
            self.activate_layer(i)
        else:
            LOG.error("no layer named: " + name)

    def deactivate_named_layer(self, name):
        if name in self.named_layers:
            i = self.named_layers[name]
            LOG.info("deactivating layer named: " + name)
            self.deactivate_layer(i)
        else:
            LOG.error("no layer named: " + name)

    def remove_named_layer(self, name):
        if name in self.named_layers:
            i = self.named_layers[name]
            LOG.info("removing layer named: " + name)
            self.remove_layer(i)
        else:
            LOG.error("no layer named: " + name)

    def replace_named_layer(self, name, intent_list=None):
        if name in self.named_layers:
            i = self.named_layers[name]
            LOG.info("replacing layer named: " + name)
            self.replace_layer(i, intent_list)
        else:
            LOG.error("no layer named: " + name)

    def replace_layer(self, layer_num, intent_list=None):
        intent_list = intent_list or []
        if self.current_layer == layer_num:
            self.deactivate_layer(layer_num)
        LOG.info("Adding layer" + str(intent_list) + " in position " + str(
            layer_num))
        self.layers[layer_num] = intent_list
        if self.current_layer == layer_num:
            self.activate_layer(layer_num)

    def remove_layer(self, layer_num):
        if layer_num >= len(self.layers):
            return False
        if self.current_layer == layer_num:
            self.deactivate_layer(layer_num)
        self.layers.pop(layer_num)
        LOG.info("Removing layer number " + str(layer_num))
        return True

    def find_layer(self, intent_name):
        layer_list = []
        for i in range(0, len(self.layers)):
            if intent_name in self.layers[i]:
                layer_list.append(i)
        return layer_list

    def disable(self):
        LOG.info("Disabling layers")
        # disable all layers
        for i in range(0, len(self.layers)):
            self.deactivate_layer(i)

    def activate_layer(self, layer_num):
        # error check
        if layer_num < 0 or layer_num > len(self.layers):
            LOG.error("invalid layer number")
            return False

        self.current_layer = layer_num

        # disable other layers
        self.disable()

        # TODO in here we should wait for all intents to be detached
        # sometimes detach intent from this step comes after register from next
        # is there some bus signal we can track?
        sleep(0.3)

        # enable layer
        LOG.info("Activating Layer " + str(layer_num))
        if layer_num < len(self.layers) and len(self.layers):
            for intent_name in self.layers[layer_num]:
                self.enable_intent(intent_name)
            return True
        return False

    def deactivate_layer(self, layer_num):
        # error check
        if layer_num < 0 or layer_num > len(self.layers):
            LOG.error("invalid layer number")
            return False
        LOG.info("Deactivating Layer " + str(layer_num))
        for intent_name in self.layers[layer_num]:
            self.disable_intent(intent_name)
        return True
